/*
    ES Knot
    Copyright 2009 Martin Storsjo

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsjo
    martin@martin.st
*/

#ifdef DESKTOP_GL_2
#define GL_OES_framebuffer_object
#define LOAD_FBO
#define PFNGLBINDFRAMEBUFFEROESPROC PFNGLBINDFRAMEBUFFEREXTPROC
#define PFNGLBINDRENDERBUFFEROESPROC PFNGLBINDRENDERBUFFEREXTPROC
#define PFNGLCHECKFRAMEBUFFERSTATUSOESPROC PFNGLCHECKFRAMEBUFFERSTATUSEXTPROC
#define PFNGLDELETERENDERBUFFERSOESPROC PFNGLDELETERENDERBUFFERSEXTPROC
#define PFNGLDELETEFRAMEBUFFERSOESPROC PFNGLDELETEFRAMEBUFFERSEXTPROC
#define PFNGLFRAMEBUFFERRENDERBUFFEROESPROC PFNGLFRAMEBUFFERRENDERBUFFEREXTPROC
#define PFNGLFRAMEBUFFERTEXTURE2DOESPROC PFNGLFRAMEBUFFERTEXTURE2DEXTPROC
#define PFNGLGENFRAMEBUFFERSOESPROC PFNGLGENFRAMEBUFFERSEXTPROC
#define PFNGLGENRENDERBUFFERSOESPROC PFNGLGENRENDERBUFFERSEXTPROC
#define PFNGLRENDERBUFFERSTORAGEOESPROC PFNGLRENDERBUFFERSTORAGEEXTPROC
#define GL_FRAMEBUFFER_BINDING_OES GL_FRAMEBUFFER_BINDING_EXT
#define GL_FRAMEBUFFER_OES GL_FRAMEBUFFER_EXT
#define GL_COLOR_ATTACHMENT0_OES GL_COLOR_ATTACHMENT0_EXT
#define GL_RENDERBUFFER_OES GL_RENDERBUFFER_EXT
#define GL_DEPTH_COMPONENT16_OES GL_DEPTH_COMPONENT16
#define GL_DEPTH_ATTACHMENT_OES GL_DEPTH_ATTACHMENT_EXT
#define GL_FRAMEBUFFER_COMPLETE_OES GL_FRAMEBUFFER_COMPLETE_EXT
#define GL_FRAMEBUFFER_UNSUPPORTED_OES GL_FRAMEBUFFER_UNSUPPORTED_EXT
#endif

#ifndef GL_OES_framebuffer_object
#define GL_FRAMEBUFFER_BINDING_OES GL_FRAMEBUFFER_BINDING
#define GL_FRAMEBUFFER_OES GL_FRAMEBUFFER
#define GL_RENDERBUFFER_OES GL_RENDERBUFFER
#define GL_COLOR_ATTACHMENT0_OES GL_COLOR_ATTACHMENT0
#define GL_RENDERBUFFER_WIDTH_OES GL_RENDERBUFFER_WIDTH
#define GL_RENDERBUFFER_HEIGHT_OES GL_RENDERBUFFER_HEIGHT
#define GL_DEPTH_COMPONENT16_OES GL_DEPTH_COMPONENT16
#define GL_DEPTH_ATTACHMENT_OES GL_DEPTH_ATTACHMENT
#define GL_FRAMEBUFFER_COMPLETE_OES GL_FRAMEBUFFER_COMPLETE
#define GL_FRAMEBUFFER_UNSUPPORTED_OES GL_FRAMEBUFFER_UNSUPPORTED

#define glBindFramebufferOES glBindFramebuffer
#define glBindRenderbufferOES glBindRenderbuffer
#define glGenFramebuffersOES glGenFramebuffers
#define glGenRenderbuffersOES glGenRenderbuffers
#define glFramebufferRenderbufferOES glFramebufferRenderbuffer
#define glGetRenderbufferParameterivOES glGetRenderbufferParameteriv
#define glRenderbufferStorageOES glRenderbufferStorage
#define glCheckFramebufferStatusOES glCheckFramebufferStatus
#define glDeleteFramebuffersOES glDeleteFramebuffers
#define glDeleteRenderbuffersOES glDeleteRenderbuffers
#define glCheckFramebufferStatusOES glCheckFramebufferStatus
#define glFramebufferTexture2DOES glFramebufferTexture2D
#endif

#if defined(USE_FBO)
#ifdef LOAD_FBO
#define glBindFramebufferOES ext_glBindFramebufferOES
#define glBindRenderbufferOES ext_glBindRenderbufferOES
#define glCheckFramebufferStatusOES ext_glCheckFramebufferStatusOES
#define glDeleteRenderbuffersOES ext_glDeleteRenderbuffersOES
#define glDeleteFramebuffersOES ext_glDeleteFramebuffersOES
#define glFramebufferRenderbufferOES ext_glFramebufferRenderbufferOES
#define glFramebufferTexture2DOES ext_glFramebufferTexture2DOES
#define glGenFramebuffersOES ext_glGenFramebuffersOES
#define glGenRenderbuffersOES ext_glGenRenderbuffersOES
#define glRenderbufferStorageOES ext_glRenderbufferStorageOES
#endif

#ifdef FBO_BINDING_STACK
int fboNumBindings = 1;
int fboBindings[10] = { 0 };
#define getFBOBinding(x) do { (x) = fboBindings[fboNumBindings - 1]; } while (0)
#define pushFBOBinding(x) do { \
		fboBindings[fboNumBindings++] = x; \
	} while (0)
#define popFBOBinding() fboNumBindings--
#else
#define getFBOBinding(x) glGetIntegerv(GL_FRAMEBUFFER_BINDING_OES, &(x))
#define pushFBOBinding(x) do { } while (0)
#define popFBOBinding() do { } while (0)
#endif

class PbufferRTT {
public:
	PbufferRTT() {
		supported = true;
#ifdef CHECK_FBO_EXT
		supported = hasExtension("GL_OES_framebuffer_object");
		if (!supported)
			return;
#endif
#ifdef LOAD_FBO
		ext_glBindFramebufferOES = (PFNGLBINDFRAMEBUFFEROESPROC) eglGetProcAddress("glBindFramebufferOES");
		ext_glBindRenderbufferOES = (PFNGLBINDRENDERBUFFEROESPROC) eglGetProcAddress("glBindRenderbufferOES");
		ext_glCheckFramebufferStatusOES = (PFNGLCHECKFRAMEBUFFERSTATUSOESPROC) eglGetProcAddress("glCheckFramebufferStatusOES");
		ext_glDeleteRenderbuffersOES = (PFNGLDELETERENDERBUFFERSOESPROC) eglGetProcAddress("glDeleteRenderbuffersOES");
		ext_glDeleteFramebuffersOES = (PFNGLDELETEFRAMEBUFFERSOESPROC) eglGetProcAddress("glDeleteFramebuffersOES");
		ext_glFramebufferRenderbufferOES = (PFNGLFRAMEBUFFERRENDERBUFFEROESPROC) eglGetProcAddress("glFramebufferRenderbufferOES");
		ext_glFramebufferTexture2DOES = (PFNGLFRAMEBUFFERTEXTURE2DOESPROC) eglGetProcAddress("glFramebufferTexture2DOES");
		ext_glGenFramebuffersOES = (PFNGLGENFRAMEBUFFERSOESPROC) eglGetProcAddress("glGenFramebuffersOES");
		ext_glGenRenderbuffersOES = (PFNGLGENRENDERBUFFERSOESPROC) eglGetProcAddress("glGenRenderbuffersOES");
		ext_glRenderbufferStorageOES = (PFNGLRENDERBUFFERSTORAGEOESPROC) eglGetProcAddress("glRenderbufferStorageOES");
		if (!ext_glBindFramebufferOES || !ext_glBindRenderbufferOES || !ext_glCheckFramebufferStatusOES || !ext_glDeleteRenderbuffersOES || !ext_glDeleteFramebuffersOES || !ext_glFramebufferRenderbufferOES || !ext_glFramebufferTexture2DOES || !ext_glGenFramebuffersOES || !ext_glGenRenderbuffersOES || !ext_glRenderbufferStorageOES) {
			supported = false;
			return;
		}
#endif
		fb = 0;
		texture = 0;
		depth = 0;
		glGenFramebuffersOES(1, &fb);
		glGenRenderbuffersOES(1, &depth);
	}
	~PbufferRTT() {
		if (!supported)
			return;
		glDeleteRenderbuffersOES(1, &depth);
		glDeleteFramebuffersOES(1, &fb);
		glDeleteTextures(1, &texture);
	}
	bool init(int w, int h) {
		if (!supported)
			return false;
#ifdef USE_OES_2
		width = w;
		height = h;
#else
		width = smallerPOT(w);
		height = smallerPOT(h);
#endif

		getFBOBinding(mainFb);
		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGB, width, height, 0, GL_RGB, GL_UNSIGNED_BYTE, 0);

		glBindFramebufferOES(GL_FRAMEBUFFER_OES, fb);
		glFramebufferTexture2DOES(GL_FRAMEBUFFER_OES, GL_COLOR_ATTACHMENT0_OES, GL_TEXTURE_2D, texture, 0);
		glBindRenderbufferOES(GL_RENDERBUFFER_OES, depth);
		glRenderbufferStorageOES(GL_RENDERBUFFER_OES, GL_DEPTH_COMPONENT16_OES, width, height);
		glFramebufferRenderbufferOES(GL_FRAMEBUFFER_OES, GL_DEPTH_ATTACHMENT_OES, GL_RENDERBUFFER_OES, depth);
		GLenum status = glCheckFramebufferStatusOES(GL_FRAMEBUFFER_OES);
		printGLErrors();
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, mainFb);
		switch (status) {
		case GL_FRAMEBUFFER_COMPLETE_OES:
			return true;
		default:
			char buf[20];
			sprintf(buf, "bad status: %d", status);
			debug(buf);
			break;
		case GL_FRAMEBUFFER_UNSUPPORTED_OES:
			debug("FBO unsupported");
			return false;
		}


		return false;
	}
	void activate() {
		getFBOBinding(mainFb);
		GLint array[4];
		glGetIntegerv(GL_VIEWPORT, array);
		mainWidth = array[2];
		mainHeight = array[3];
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, fb);
		glViewport(0, 0, width, height);
		pushFBOBinding(fb);
	}
	void finish() {
		glBindFramebufferOES(GL_FRAMEBUFFER_OES, mainFb);
		popFBOBinding();
		glViewport(0, 0, mainWidth, mainHeight);
	}
	void bind() {
		glBindTexture(GL_TEXTURE_2D, texture);
	}
	int getWidth() {
		return width;
	}
	int getHeight() {
		return height;
	}
	void clear() {
		activate();
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);
		finish();
	}

private:
	bool supported;
	GLuint fb;
	GLuint color;
	GLuint depth;
	GLuint texture;
	GLint mainFb;

	int mainWidth, mainHeight;
	int width, height;

#ifdef LOAD_FBO
	PFNGLBINDFRAMEBUFFEROESPROC ext_glBindFramebufferOES;
	PFNGLBINDRENDERBUFFEROESPROC ext_glBindRenderbufferOES;
	PFNGLCHECKFRAMEBUFFERSTATUSOESPROC ext_glCheckFramebufferStatusOES;
	PFNGLDELETERENDERBUFFERSOESPROC ext_glDeleteRenderbuffersOES;
	PFNGLDELETEFRAMEBUFFERSOESPROC ext_glDeleteFramebuffersOES;
	PFNGLFRAMEBUFFERRENDERBUFFEROESPROC ext_glFramebufferRenderbufferOES;
	PFNGLFRAMEBUFFERTEXTURE2DOESPROC ext_glFramebufferTexture2DOES;
	PFNGLGENFRAMEBUFFERSOESPROC ext_glGenFramebuffersOES;
	PFNGLGENRENDERBUFFERSOESPROC ext_glGenRenderbuffersOES;
	PFNGLRENDERBUFFERSTORAGEOESPROC ext_glRenderbufferStorageOES;
#endif
};
#elif defined(USE_PBUFFER)
class PbufferRTT {
public:
	PbufferRTT() {
		display = EGL_NO_DISPLAY;
		surface = EGL_NO_SURFACE;
		texture = 0;
	}
	~PbufferRTT() {
		if (surface != EGL_NO_SURFACE)
			eglDestroySurface(display, surface);
		glDeleteTextures(1, &texture);
	}
	bool init(int w, int h) {
		display = eglGetCurrentDisplay();
		context = eglGetCurrentContext();
		EGLint attribList[] = {
			EGL_DEPTH_SIZE, 16,
			EGL_SURFACE_TYPE, EGL_PBUFFER_BIT | EGL_WINDOW_BIT,
			EGL_BIND_TO_TEXTURE_RGB, EGL_TRUE,
#ifdef USE_OES_2
			EGL_RENDERABLE_TYPE, EGL_OPENGL_ES2_BIT,
#endif
			EGL_NONE
		};

		EGLConfig config;
		EGLint numConfigs;
		if (eglChooseConfig(display, attribList, &config, 1, &numConfigs) == EGL_FALSE) {
			return false;
		}
		if (numConfigs < 1) {
			return false;
		}
		EGLint surfaceAttribs[] = {
			EGL_WIDTH, 256,
			EGL_HEIGHT, 256,
			EGL_TEXTURE_FORMAT, EGL_TEXTURE_RGB,
			EGL_TEXTURE_TARGET, EGL_TEXTURE_2D,
			EGL_MIPMAP_TEXTURE, EGL_FALSE,
			EGL_NONE
		};

		surface = eglCreatePbufferSurface(display, config, surfaceAttribs);
		if (surface == EGL_NO_SURFACE) {
			return false;
		}
		eglQuerySurface(display, surface, EGL_WIDTH, &width);
		eglQuerySurface(display, surface, EGL_HEIGHT, &height);
		glGenTextures(1, &texture);
		glBindTexture(GL_TEXTURE_2D, texture);
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		return true;
	}
	void activate() {
		eglReleaseTexImage(display, surface, EGL_BACK_BUFFER);

		prevDrawSurface = eglGetCurrentSurface(EGL_DRAW);
		prevReadSurface = eglGetCurrentSurface(EGL_READ);

		eglMakeCurrent(display, surface, surface, context);
		glViewport(0, 0, width, height);
	}
	void finish() {
		glBindTexture(GL_TEXTURE_2D, texture);
		eglBindTexImage(display, surface, EGL_BACK_BUFFER);

		eglMakeCurrent(display, prevDrawSurface, prevReadSurface, context);
		EGLint prevWidth, prevHeight;
		eglQuerySurface(display, prevDrawSurface, EGL_WIDTH, &prevWidth);
		eglQuerySurface(display, prevDrawSurface, EGL_HEIGHT, &prevHeight);
		glViewport(0, 0, prevWidth, prevHeight);
	}
	void bind() {
		glBindTexture(GL_TEXTURE_2D, texture);
	}
	int getWidth() {
		return width;
	}
	int getHeight() {
		return height;
	}
	void clear() {
		activate();
		glClearColor(0, 0, 0, 0);
		glClear(GL_COLOR_BUFFER_BIT);
		finish();
	}

private:
	EGLDisplay display;
	EGLSurface surface;
	EGLContext context;

	EGLSurface prevDrawSurface;
	EGLSurface prevReadSurface;
	GLuint texture;
	EGLint width, height;
};
#else
class PbufferRTT {
public:
	bool init(int width, int height) {
		return false;
	}
	void activate() { }
	void finish() { }
	void bind() { }
	void clear() { }
	int getWidth() { return 0; }
	int getHeight() { return 0; }
};
#endif

