package st.martin.esknot;

import javax.microedition.khronos.egl.EGL10;
import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.egl.EGLContext;
import javax.microedition.khronos.egl.EGLDisplay;

import android.opengl.GLSurfaceView;

public class GLES2EGLContextFactory implements GLSurfaceView.EGLContextFactory, GLSurfaceView.EGLConfigChooser {

	public static final int EGL_CONTEXT_CLIENT_VERSION = 0x3098;
	public static final int EGL_OPENGL_ES2_BIT = 4;

	public EGLContext createContext(EGL10 egl, EGLDisplay display, EGLConfig eglConfig) {
		int[] attribs = new int[3];
		attribs[0] = EGL_CONTEXT_CLIENT_VERSION;
		attribs[1] = 2;
		attribs[2] = EGL10.EGL_NONE;
		return egl.eglCreateContext(display, eglConfig, EGL10.EGL_NO_CONTEXT, useGl2 ? attribs : null);
	}

	public void destroyContext(EGL10 egl, EGLDisplay display, EGLContext context) {
		egl.eglDestroyContext(display, context);
	}

	private int[] getAttribs(boolean gl2) {
		int[] attribs = new int[gl2 ? 5 : 3];
		int i = 0;
		if (gl2) {
			attribs[i++] = EGL10.EGL_RENDERABLE_TYPE;
			attribs[i++] = EGL_OPENGL_ES2_BIT;
		}
		attribs[i++] = EGL10.EGL_DEPTH_SIZE;
		attribs[i++] = 16;
		attribs[i++] = EGL10.EGL_NONE;
		return attribs;
	}

	public EGLConfig chooseConfig(EGL10 egl, EGLDisplay display) {
		int[] attribs = getAttribs(true);
		int[] numConfigs = new int[1];
		egl.eglChooseConfig(display, attribs, null, 0, numConfigs);
		if (numConfigs[0] <= 0) {
			attribs = getAttribs(false);
			egl.eglChooseConfig(display, attribs, null, 0, numConfigs);
			KnotRenderer.setGlesVersion(1);
			useGl2 = false;
		} else {
			KnotRenderer.setGlesVersion(2);
		}
		EGLConfig[] configs = new EGLConfig[numConfigs[0]];
		egl.eglChooseConfig(display, attribs, configs, numConfigs[0], numConfigs);
		return configs[0];
	}

	private boolean useGl2 = true;

}
