package st.martin.esknot;

import javax.microedition.khronos.egl.EGLConfig;
import javax.microedition.khronos.opengles.GL10;

import android.view.MotionEvent;

public class KnotRenderer implements android.opengl.GLSurfaceView.Renderer {
	private static int glesVersion = 1;
	private static boolean isLoaded = false;

	public static void setGlesVersion(int version) {
		glesVersion = version;
	}

	public static void load() {
		if (isLoaded)
			return;
		try {
			if (glesVersion == 2) {
				System.loadLibrary("esknot2");
				isLoaded = true;
			} else {
				throw new Exception("choose GLES 1");
			}
		} catch (Throwable t) {
			t.printStackTrace();
			System.loadLibrary("esknot");
			glesVersion = 1;
			isLoaded = true;
		}
	}

	public void onDrawFrame(GL10 gl) {
		if (diffX != 0 || diffY != 0) {
			rotate(diffX, diffY);
			diffX = diffY = 0;
		}
		if (texModeChanged) {
			setTexMode(texMode);
			texModeChanged = false;
		}
		if (vertexUpdatingChanged) {
			setUpdateVertices(vertexUpdating);
			vertexUpdatingChanged = false;
		}
		if (radialBlurChanged) {
			setRadialBlur(radialBlur);
			radialBlurChanged = false;
		}
		if (motionBlurChanged) {
			setMotionBlur(motionBlur);
			motionBlurChanged = false;
		}
		nativeOnDrawFrame(System.currentTimeMillis());
	}

	public void onSurfaceChanged(GL10 gl, int width, int height) {
		load();
		nativeOnSurfaceChanged(width, height);
		if (!getMatrixSupported())
			hasGravity = false;
		setTexMode(texMode);
		setUpdateVertices(vertexUpdating);
		setRadialBlur(radialBlur);
		setMotionBlur(motionBlur);
		enableGravity(hasGravity);
		enableRotate(!hasGravity);
		texModeChanged = false;
		vertexUpdatingChanged = false;
		radialBlurChanged = false;
		motionBlurChanged = false;
	}

	public void onSurfaceCreated(GL10 gl, EGLConfig config) {
		load();
		nativeOnSurfaceCreated();
	}
	
	private native void nativeOnDrawFrame(long time);
	private native void nativeOnSurfaceCreated();
	private native void nativeOnSurfaceChanged(int width, int height);
	
	public boolean onTouchEvent(MotionEvent event) {
		if (event.getAction() == MotionEvent.ACTION_DOWN) {
			enableRotate(false);
			enableGravity(false);
			distance = 0;
			lastX = (int) event.getX();
			lastY = (int) event.getY();
		} else if (event.getAction() == MotionEvent.ACTION_MOVE) {
			int curX = (int) event.getX();
			int curY = (int) event.getY();
			int curDiffX = curX - lastX;
			int curDiffY = curY - lastY;
			diffX += curDiffX;
			diffY += curDiffY;
			lastX = curX;
			lastY = curY;
			distance += Math.abs(curDiffX) + Math.abs(curDiffY);
		} else if (event.getAction() == MotionEvent.ACTION_UP) {
			if (hasGravity)
				enableGravity(true);
			else
				enableRotate(true);
			if (distance < 25)
				startMorph();
		}
		return true;
	}
	private native void enableRotate(boolean rotate);
	private native void rotate(int x, int y);
	private native void startMorph();
	private native void enableGravity(boolean gravity);
	private native void setGravity(float x, float y, float z);
	private native boolean getMatrixSupported();
	
	public void toggleTextureMode() {
		texMode = (texMode + 1) % 4;
		if (texMode == 3) {
			/* Skip celshading if it isn't supported */
			if (isLoaded && !getMatrixSupported())
				texMode = 0;
		}
		texModeChanged = true;
	}
	
	private native void setTexMode(int mode);
	private native void setUpdateVertices(boolean update);
	private native void setRadialBlur(int mode);
	private native void setMotionBlur(int mode);

	public void toggleVertexUpdating() {
		vertexUpdating = !vertexUpdating;
		vertexUpdatingChanged = true;
	}

	public void toggleRadialBlur() {
		radialBlur = (radialBlur + 1) % 3;
		radialBlurChanged = true;
	}

	public void toggleMotionBlur() {
		motionBlur = (motionBlur + 1) % 2;
		motionBlurChanged = true;
	}
	
	public void setHasGravity(boolean gravity) {
		hasGravity = gravity;
		if (!isLoaded)
			return;
		if (!getMatrixSupported())
			hasGravity = false;
		enableGravity(hasGravity);
		enableRotate(!hasGravity);
	}
	
	public void setGravityVector(float x, float y, float z) {
		if (!isLoaded)
			return;
		setGravity(x, y, z);
	}

	private long privData;
	
	private int distance;
	private int lastX, lastY;
	private int diffX, diffY;
	
	private int texMode = 2;
	private boolean texModeChanged;

	private boolean vertexUpdating = true;
	private boolean vertexUpdatingChanged;

	private int radialBlur = 0;
	private boolean radialBlurChanged;

	private int motionBlur = 0;
	private boolean motionBlurChanged;
	
	private boolean hasGravity = false;

	private native void close();
	public void finalize() {
		close();
	}
}
